from datetime import datetime

from google_internal_apis import LibraryServiceRpc


class LibraryService(LibraryServiceRpc):
    async def add_tags(self, book_ids, tag_id):
        return await super().add_tags(
            [
                [
                    [book_id, tag_id, str(int(datetime.now().timestamp() * 1000))]
                    for book_id in book_ids
                ]
            ]
        )

    async def list_tags(self):
        [tags, tagged] = await super().list_tags()

        return {
            "tags": {name: tag_id for name, tag_id, *_ in tags},
            "tagged": [
                {
                    "book_id": book_id,
                    "tag_id": tag_id,
                    "tagged_at": datetime.fromtimestamp(int(tagged_at) / 1000),
                }
                for book_id, tag_id, tagged_at, *_ in tagged
            ],
        }



import os
from mimetypes import guess_type

import requests
from googleapiclient.discovery import Resource

from .const import COOKIE_TXT


def steal_cookie(data: dict) -> str | None:
    events = data["events"]
    events = [event["params"] for event in events if event["type"] == 201]

    for event in events:
        headers = event["headers"]
        headers = dict(header.split(": ", 1) for header in headers)

        authority = headers.get(":authority")
        if authority and "playbooks" in authority:
            return headers["cookie"]
    return None


def upload_with_scotty(books: Resource, filename: str) -> dict:
    content_id = resume_upload(filename)

    return (
        books.cloudloading()
        .addBook(
            upload_client_token=content_id,
        )
        .execute()
    )


def start_upload(session: requests.Session, filename: str, mimetype: str):
    stat = os.stat(filename)
    filesize = stat.st_size
    title = os.path.basename(filename)

    res = session.post(
        "https://docs.google.com/upload/books/library/upload",
        params={"authuser": "0", "opi": "113040485"},
        headers={
            "accept": "*/*",
            # "content-length": "557",
            "origin": "https://docs.google.com",
            "priority": "u=1, i",
            "referer": "https://docs.google.com/",
            "sec-fetch-dest": "empty",
            "sec-fetch-mode": "cors",
            "sec-fetch-site": "same-origin",
            "x-goog-upload-command": "start",
            "x-goog-upload-header-content-length": str(filesize),
            "x-goog-upload-header-content-type": mimetype,
            "x-goog-upload-protocol": "resumable",
        },
        json={
            "protocolVersion": "0.8",
            "createSessionRequest": {
                "fields": [
                    {
                        "external": {
                            "name": "file",
                            "filename": title,
                            "put": {},
                            "size": filesize,
                        }
                    },
                    {
                        "inlined": {
                            "name": "title",
                            "content": title,
                            "contentType": "text/plain",
                        }
                    },
                    {
                        "inlined": {
                            "name": "addtime",
                            "content":
                            # int(stat.st_atime * 1000),
                            "1723462164781",
                            "contentType": "text/plain",
                        }
                    },
                    {
                        "inlined": {
                            "name": "onepick_version",
                            "content": "v2",
                            "contentType": "text/plain",
                        }
                    },
                    {
                        "inlined": {
                            "name": "onepick_host_id",
                            "content": "20",
                            "contentType": "text/plain",
                        }
                    },
                    {
                        "inlined": {
                            "name": "onepick_host_usecase",
                            "content": "PlayBooks.Web",
                            "contentType": "text/plain",
                        }
                    },
                ]
            },
        },
    )
    """
    X-Goog-Upload-Chunk-Granularity:
    262144
    X-Goog-Upload-Control-Url:
    https://docs.google.com/upload/books/library/upload?authuser=0&opi=113040485&upload_id=AHxI1nMkzatj-c5HRv1gmlu1-AME4SLqMPzCKrBPPvMJJtcDHtbrY-MyVI9q84dRBRG-GStp0MKj_bhLYiXyvfPlHGpWKQdaKJeKqPtHutcsL8Qc-Q&upload_protocol=resumable
    X-Goog-Upload-Status:
    active
    X-Goog-Upload-Url:
    https://docs.google.com/upload/books/library/upload?authuser=0&opi=113040485&upload_id=AHxI1nMkzatj-c5HRv1gmlu1-AME4SLqMPzCKrBPPvMJJtcDHtbrY-MyVI9q84dRBRG-GStp0MKj_bhLYiXyvfPlHGpWKQdaKJeKqPtHutcsL8Qc-Q&upload_protocol=resumable
    X-Guploader-Uploadid:
    AHxI1nMkzatj-c5HRv1gmlu1-AME4SLqMPzCKrBPPvMJJtcDHtbrY-MyVI9q84dRBRG-GStp0MKj_bhLYiXyvfPlHGpWKQdaKJeKqPtHutcsL8Qc-Q
    """
    if not res.ok or "x-goog-upload-url" not in res.headers:
        raise Exception(res.text)

    return res.headers["X-Goog-Upload-Url"]


def resume_upload(filename):
    session = requests.Session()
    session.headers.update(
        {
            "User-Agent": (
                "Mozilla/5.0 (Windows NT 10.0; Win64; x64) "
                "AppleWebKit/537.36 (KHTML, like Gecko) "
                "Chrome/126.0.0.0 Safari/537.36"
            ),
            "cookie": COOKIE_TXT.read_text(),
        }
    )

    mimetype = guess_type(filename)[0]
    if not mimetype:
        raise Exception(f"Could not determine mimetype for {filename}")

    url = start_upload(session, filename, mimetype)

    res = session.put(
        url,
        headers={
            "accept": "*/*",
            "content-type": mimetype,
            "origin": "https://docs.google.com",
            "priority": "u=1, i",
            "sec-fetch-dest": "empty",
            "sec-fetch-mode": "cors",
            "sec-fetch-site": "same-origin",
            "x-goog-upload-command": "upload, finalize",
            "x-goog-upload-offset": "0",
            "referer": "https://docs.google.com/",
        },
        data=open(filename, "rb").read(),
    )

    js = res.json()
    if "errorMessage" in js:
        raise Exception(js["errorMessage"])

    completion_info = js["sessionStatus"]["additionalInfo"][
        "uploader_service.GoogleRupioAdditionalInfo"
    ]["completionInfo"]

    assert completion_info["status"] == "SUCCESS"

    return completion_info["customerSpecificInfo"]["contentId"]



from mimetypes import guess_type
from os.path import basename, splitext

from googleapiclient.discovery import Resource


def upload_with_drive(drive: Resource, books: Resource, filename: str) -> dict:
    name = splitext(basename(filename))[0]
    mime_type = guess_type(filename)[0]
    response = (
        drive.files().create(media_body=filename, media_mime_type=mime_type).execute()
    )

    return (
        books.cloudloading()
        .addBook(
            # A drive document id. The upload_client_token must not be set.
            drive_document_id=response["id"],
            # The document MIME type.
            # It can be set only if the drive_document_id is set.
            mime_type=mime_type,
            # The document name.
            # It can be set only if the drive_document_id is set.
            name=name,
        )
        .execute()
    )



#!/usr/bin/env python3
import json
import logging
import time
from functools import wraps
from json import JSONDecodeError
from mimetypes import add_type
from pathlib import Path
from typing import Callable, Optional, TypeVar

import httplib2
import rich_click as click
import uvloop
from click.exceptions import Abort, BadParameter
from click.globals import get_current_context
from googleapiclient.discovery import Resource, build
from oauth2client.client import flow_from_clientsecrets
from oauth2client.file import Storage
from oauth2client.tools import argparser, run_flow
from rich.logging import RichHandler

import google_internal_apis as endpoints_rpc
from google_internal_apis import get_client
from google_internal_apis.ghunter import RpcService

from .const import COOKIE_TXT, PATH
from .drive import upload_with_drive
from .endpoints import LibraryService
from .scotty import steal_cookie, upload_with_scotty

logging.basicConfig(handlers=[RichHandler(rich_tracebacks=True)])

add_type("application/epub+zip", ".epub")


def get_http():
    assert argparser
    args = argparser.parse_args(["--noauth_local_webserver"])
    storage = Storage(PATH / "credentials.json")
    credentials = storage.get()
    if credentials is None:
        flow = flow_from_clientsecrets(
            PATH / "client_secrets.json",
            scope=[
                "https://www.googleapis.com/auth/drive.file",
                "openid",
                "https://www.googleapis.com/auth/drive",
                "https://www.googleapis.com/auth/books",
            ],
        )
        credentials = run_flow(flags=args, flow=flow, storage=storage)
    http = credentials.authorize(httplib2.Http())
    if credentials.access_token_expired:
        credentials.refresh(http)
    return http


def get_volume(
    books: Resource, name: Optional[str] = None, volumeId: Optional[str] = None
):
    return next(
        volume
        for volume in books.volumes().useruploaded().list().execute()["items"]
        if (volume["volumeInfo"]["title"] == name or volume["id"] == volumeId)
    )


def paginate(method: Callable, *args, **kwargs):
    request = method(*args, **kwargs)
    start_index = 0
    while request:
        res = request.execute()
        if "items" not in res:
            break
        yield from res["items"]
        start_index += len(res["items"])
        request = method(*args, startIndex=start_index, **kwargs)


@click.group()
def main() -> None:
    pass


def verbose_flag(func):
    @click.option("--verbose", is_flag=True)
    @wraps(func)
    def wrapper(*args, **kwargs):
        verbose = kwargs.pop("verbose")
        logging.getLogger().setLevel(logging.DEBUG if verbose else logging.INFO)
        try:
            return func(*args, **kwargs)
        except BaseException as e:
            logging.exception(e)
            raise Abort(e)

    return wrapper


@main.command()
@click.argument(
    "files",
    required=True,
    nargs=-1,
    type=click.Path(exists=True, readable=True, path_type=Path, dir_okay=False),
)
@click.option("--use-drive", is_flag=True)
@click.option("--bookshelf", help="Add the uploaded books to this bookshelf")
@verbose_flag
def upload(files: list[str], use_drive: bool, bookshelf: str):
    """
    Upload files to Google Books
    """

    http = get_http()

    books = build("books", "v1", http=http)
    drive = build("drive", "v3", http=http)

    uploads = [
        upload_with_drive(drive, books, filename)
        if use_drive
        else upload_with_scotty(books, filename)
        for filename in files
    ]
    for upl in uploads:
        monitor(books, upl["volumeId"])

    if not bookshelf:
        return

    uvloop.run(add_multiple_to_shelf([upl["volumeId"] for upl in uploads], bookshelf))


async def add_multiple_to_shelf(book_ids: list[str], shelf_name: str):
    service = await get_client(LibraryService)

    tag_id = await get_shelf(service, shelf_name)

    print(await service.add_tags(book_ids, tag_id))


def load_json(ctx, param, filename):
    try:
        with open(filename) as f:
            return json.load(f)
    except JSONDecodeError as e:
        raise BadParameter(e) from e


def asyncio(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        return uvloop.run(func(*args, **kwargs))

    return wrapper


@main.command()
@click.argument(
    "filename",
    type=click.Path(exists=True, readable=True, path_type=Path, dir_okay=False),
    callback=load_json,
)
@verbose_flag
def steal(data: dict):
    """
    Steal the cookie from a Chrome net-export log
    """
    cookie = steal_cookie(data)
    if not cookie:
        raise Exception("Could not find cookie")
    PATH.mkdir(parents=True, exist_ok=True)
    COOKIE_TXT.write_text(cookie)


@main.group()
def shelves():
    pass


async def get_shelf(service: LibraryService, shelf_name: str):
    tags = await service.list_tags()

    if shelf_name not in tags["tags"]:
        ctx = get_current_context()
        param = next(p for p in ctx.command.params if p.name == "bookshelf")
        raise BadParameter(f'Shelf with name "{shelf_name}" not found', param=param)

    return tags["tags"][shelf_name]


T = TypeVar("T", bound=RpcService)


@shelves.command("add", help="add book to shelf")
@click.argument("book_id")
@click.argument("bookshelf")
@verbose_flag
@asyncio
async def add_to_shelf(book_id: str, bookshelf: str):
    service = await get_client(LibraryService)

    tag_id = await get_shelf(service, bookshelf)

    print(await service.add_tags([book_id], tag_id))


@shelves.command("list", help="list shelves")
@verbose_flag
@asyncio
async def list_shelves():
    service = await get_client(LibraryService)

    tags = await service.list_tags()

    for name in tags["tags"]:
        print(name)


def validate_method(ctx, param, value):
    service = ctx.params["service"]

    stype = click.Choice([k for k, v in vars(service).items() if callable(v)])

    return stype(value)


@main.command()
@click.argument(
    "service",
    type=click.Choice(
        [
            k
            for k, v in vars(endpoints_rpc).items()
            if isinstance(v, type) and issubclass(v, RpcService) and v != RpcService
        ]
    ),
    callback=lambda ctx, param, value: getattr(endpoints_rpc, value),
)
@click.argument("method", callback=validate_method)
@click.argument("data", required=False)
@verbose_flag
@asyncio
async def rpc(service: type[RpcService], method: str, data: Optional[str]):
    service = await get_client(service)
    bound_method = getattr(service, method)
    logging.info(
        "tags: %s", await (bound_method() if data is None else bound_method(data))
    )


def monitor(books: Resource, volume_id: str) -> None:
    wait = 1.0

    while True:
        state = get_volume(books, volumeId=volume_id)["userInfo"][
            "userUploadedVolumeInfo"
        ]["processingState"]
        print(state)
        if state.startswith("COMPLETED_"):
            break
        time.sleep(wait)
        wait *= 1.5


if __name__ == "__main__":
    main()
